{* Unit for evaluations with real numbers and selectable precision
   v. 0.9 for FPC

   Copyright (C) 1999 Franco Milani (franco@intercity.it).

   You can use or modify this program for personal purposes. You can also
   redistribute it, but only in the original unmodified form. The possible
   changes in the distributed versions will be made by the author, also
   according to hints, requests or bug reports. 

   This program is released in the hope that it will be useful, but on an
   as-is basis. No warranty of any kind is offered, either expressed or
   implied, including, without limitation, the implied warranties of
   merchantability and fitness for a particular purpose. *}

{$H+}                   { to use ansistrings in strtox and xtostr procedures }
UNIT mpcalc;

INTERFACE

CONST

maxbuf = 1024;

TYPE

bufnum = array[1..maxbuf] of byte;

mp = record                                      { the multi precision number }
c : bufnum;
s : boolean;
n : word;
i : word;
d : word;
end;

ex = extended;

{* procedures and functions *}

procedure xprecision (n: word);               { to set the digits number }
function  xdigits: word;                      { returns the current precision }
function  xerror: byte;                       { error control function }
procedure xdis (a: mp);                       { displays a mp on the screen }
procedure xapp (a: mp; var b: mp; w:word); { approximates a at the digit number
                                             w and stores the result in b }
procedure strtox (s: string; var a: mp; var w: word); { converts an ansistring
                                                      to mp. If the conversion
                                                      is correct w=0, otherwise
                                                      w contains the number of
                                                      the character where the
                                                      first error occurs }
procedure xtostr (a: mp; var s: string);      { converts an mp to ansistring }

{* overloaded operators *}

{ comparison }

operator =  (a,b: mp) x: boolean;             { a (mp) = b (mp) }
operator >  (a,b: mp) x: boolean;             { a (mp) > b (mp) }
operator <  (a,b: mp) x: boolean;             { a (mp) < b (mp) }
operator >= (a,b: mp) x: boolean;             { a (mp) >= b (mp) }
operator <= (a,b: mp) x: boolean;             { a (mp) <= b (mp) }

{ assignments }

operator := (e: ex) a: mp;                    { a (mp) := e (extended) }
operator := (a: mp) e: ex;                    { e (extended) := a (mp) }

{ additive inverse }

operator -  (a: mp) b: mp;                    { - a (mp) }

{ arithmetic operations }

operator +  (a,b: mp) c: mp;                  { a (mp) + b (mp) -> mp } 
operator +  (a: mp; e: ex) b: mp;             { a (mp) + e (extended) -> mp }
operator +  (e: ex; a: mp) b: mp;             { e (extended) + a (mp) -> mp }

operator -  (a,b: mp) c: mp;                  { a (mp) - b (mp) -> mp }
operator -  (a: mp; e: ex) b: mp;             { a (mp) - e (extended) -> mp }
operator -  (e: ex; a: mp) b: mp;             { e (extended) - a (mp) -> mp }

operator *  (a,b: mp) c: mp;                  { a (mp) * b (mp) -> mp }
operator *  (a: mp; e: ex) b: mp;             { a (mp) * e (extended) -> mp }
operator *  (e: ex; a: mp) b: mp;             { e (extended) * a (mp) -> mp }

operator /  (a,b: mp) c: mp;                  { a (mp) / b (mp) -> mp }
operator /  (a: mp; e: ex) b: mp;             { a (mp) / e (extended) -> mp }
operator /  (e: ex; a: mp) b: mp;             { e (extended) / a (mp) -> mp }

operator ** (a,b: mp) c: mp;                  { a (mp) ** b (mp) -> mp }
operator ** (e: ex; a: mp) b: mp;             { e (extended) ** a (mp) -> mp }

{* pi and Euler's number *}

function mpi: mp;                             { returns the constant pi }
function meu: mp;                             { returns the Euler's number }

{* mathematical functions, applied to mp and extended numbers.
'er' are the possible values returned by the function xerror *}

function mabs   (a: mp) : mp; {absolute value}
function mabs   (e: ex) : mp;
function mint   (a: mp) : mp; {integer part}
function mint   (e: ex) : mp;
function mfrac  (a: mp) : mp; {fractional part}
function mfrac  (e: ex) : mp;
function msign  (a: mp) : mp; {sign function: -1 if a<0, 0 if a=0, +1 if a>0}
function msign  (e: ex) : mp;
function mceil  (a: mp) : mp; {the smallest integer >=a}
function mceil  (e: ex) : mp;
function mfloor (a: mp) : mp; {the largest integer <=a}
function mfloor (e: ex) : mp;
function msqr   (a: mp) : mp; {square, if overflow er=7}
function msqr   (e: ex) : mp;
function msqrt  (a: mp) : mp; {square root, if a<0 er=10}
function msqrt  (e: ex) : mp;
function mfact  (a: mp) : mp; {factorial, er=33 if a incorrect, 34 if overflow}
function mfact  (e: ex) : mp;
function mexp   (a: mp) : mp; {exponential, er=19 if overflow}
function mexp   (e: ex) : mp;
function mln    (a: mp) : mp; {natural logarithm, er=20 if a<=0}
function mln    (e: ex) : mp;
function mlog   (a: mp) : mp; {base 10 logarithm, er=21 if a<=0}
function mlog   (e: ex) : mp;
function msin   (a: mp) : mp; {sine}
function msin   (e: ex) : mp;
function mcos   (a: mp) : mp; {cosine}
function mcos   (e: ex) : mp;
function mtan   (a: mp) : mp; {tangent, er=11 if overflow}
function mtan   (e: ex) : mp;
function mcot   (a: mp) : mp; {cotangent, er=12 if overflow}
function mcot   (e: ex) : mp;
function msec   (a: mp) : mp; {secant, er=13 if overflow}
function msec   (e: ex) : mp;
function mcsc   (a: mp) : mp; {cosecant, er=14 if overflow}
function mcsc   (e: ex) : mp;
function masin  (a: mp) : mp; {arcsine, er=15 if a<-1 or a>1}
function masin  (e: ex) : mp;
function macos  (a: mp) : mp; {arccosine, er=16 if a<-1 or a>1}
function macos  (e: ex) : mp;
function matan  (a: mp) : mp; {arctangent}
function matan  (e: ex) : mp;
function macot  (a: mp) : mp; {arccotangent}
function macot  (e: ex) : mp;
function masec  (a: mp) : mp; {arcsecant, er=17 if a>-1 and a<1}
function masec  (e: ex) : mp;
function macsc  (a: mp) : mp; {arccosecant, er=18 if a>-1 and a<1}
function macsc  (e: ex) : mp;
function msinh  (a: mp) : mp; {hyperbolic sine, er=24 if overflow}
function msinh  (e: ex) : mp;
function mcosh  (a: mp) : mp; {hyperbolic cosine, er=25 if overflow}
function mcosh  (e: ex) : mp;
function mtanh  (a: mp) : mp; {hyperbolic tangent}
function mtanh  (e: ex) : mp;
function mcoth  (a: mp) : mp; {hyperbolic cotangent, er=26 if a=0}
function mcoth  (e: ex) : mp;
function msech  (a: mp) : mp; {hyperbolic secant}
function msech  (e: ex) : mp;
function mcsch  (a: mp) : mp; {hyperbolic cosecant, er=27 if a=0}
function mcsch  (e: ex) : mp;
function masinh (a: mp) : mp; {inv. hyp. sine}
function masinh (e: ex) : mp;
function macosh (a: mp) : mp; {inv. hyp. cosine, er=28 if a<1}
function macosh (e: ex) : mp;
function matanh (a: mp) : mp; {inv. hyp. tangent, er=29 if a<=-1 or a>=1}
function matanh (e: ex) : mp;
function macoth (a: mp) : mp; {inv. hyp. cotangent, er=30 if a>=-1 and a<=1}
function macoth (e: ex) : mp;
function masech (a: mp) : mp; {inv. hyp. secant, er=31 if a<=0 or a>1}
function masech (e: ex) : mp;
function macsch (a: mp) : mp; {inv. hyp. cosecant, er=32 if a=0}
function macsch (e: ex) : mp;

IMPLEMENTATION

CONST

maxdig = 500;
vcc : comp = 100000000;
dpot : array[0..4] of word = (1,10,100,1000,10000);
d10 : array[0..99] of byte =
(0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,
3,3,4,4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,6,6,6,6,6,6,6,6,6,6,7,7,7,7,7,7,7,
7,7,7,8,8,8,8,8,8,8,8,8,8,9,9,9,9,9,9,9,9,9,9);
m10 : array[0..99] of byte =
(0,1,2,3,4,5,6,7,8,9,0,1,2,3,4,5,6,7,8,9,0,1,2,3,4,5,6,7,8,9,0,1,2,3,4,5,6,7,
8,9,0,1,2,3,4,5,6,7,8,9,0,1,2,3,4,5,6,7,8,9,0,1,2,3,4,5,6,7,8,9,0,1,2,3,4,5,6,
7,8,9,0,1,2,3,4,5,6,7,8,9,0,1,2,3,4,5,6,7,8,9);
ck : array[0..19] of byte = (0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1);
cd : array[-10..9] of byte = (0,1,2,3,4,5,6,7,8,9,0,1,2,3,4,5,6,7,8,9);
cu : array[-10..9] of byte = (1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0);
tr : array[0..99] of word =
(0,1,1,1,2,2,2,2,2,3,3,3,3,3,3,3,4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,
6,6,6,6,6,6,6,6,6,6,6,6,6,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,8,8,8,8,8,8,8,8,8,
8,8,8,8,8,8,8,8,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9);
p7 : longint = 10000000;
p6 : longint = 1000000;
p5 : longint = 100000;
p4 : longint = 10000;
p3 : longint = 1000;
p2 : longint = 100;
p1 : longint = 10;
pow10 : array[1..18] of comp =
(10.0,100.0,1000.0,10000.0,100000.0,1000000.0,10000000.0,100000000.0,
1000000000.0,10000000000.0,100000000000.0,1000000000000.0,10000000000000.0,
100000000000000.0,1000000000000000.0,10000000000000000.0,100000000000000000.0,
1000000000000000000.0);
maxc : array[1..19] of byte = (9,2,2,3,3,7,2,0,3,6,8,5,4,7,7,5,8,0,7);
maxl : array[1..10] of byte = (2,1,4,7,4,8,3,6,4,7);
xpg: array[1..500] of byte =
(3,1,4,1,5,9,2,6,5,3,5,8,9,7,9,3,2,3,8,4,6,2,6,4,3,3,8,3,2,7,9,5,0,2,8,8,4,1,
9,7,1,6,9,3,9,9,3,7,5,1,0,5,8,2,0,9,7,4,9,4,4,5,9,2,3,0,7,8,1,6,4,0,6,2,8,6,
2,0,8,9,9,8,6,2,8,0,3,4,8,2,5,3,4,2,1,1,7,0,6,7,9,8,2,1,4,8,0,8,6,5,1,3,2,8,
2,3,0,6,6,4,7,0,9,3,8,4,4,6,0,9,5,5,0,5,8,2,2,3,1,7,2,5,3,5,9,4,0,8,1,2,8,4,
8,1,1,1,7,4,5,0,2,8,4,1,0,2,7,0,1,9,3,8,5,2,1,1,0,5,5,5,9,6,4,4,6,2,2,9,4,8,
9,5,4,9,3,0,3,8,1,9,6,4,4,2,8,8,1,0,9,7,5,6,6,5,9,3,3,4,4,6,1,2,8,4,7,5,6,4,
8,2,3,3,7,8,6,7,8,3,1,6,5,2,7,1,2,0,1,9,0,9,1,4,5,6,4,8,5,6,6,9,2,3,4,6,0,3,
4,8,6,1,0,4,5,4,3,2,6,6,4,8,2,1,3,3,9,3,6,0,7,2,6,0,2,4,9,1,4,1,2,7,3,7,2,4,
5,8,7,0,0,6,6,0,6,3,1,5,5,8,8,1,7,4,8,8,1,5,2,0,9,2,0,9,6,2,8,2,9,2,5,4,0,9,
1,7,1,5,3,6,4,3,6,7,8,9,2,5,9,0,3,6,0,0,1,1,3,3,0,5,3,0,5,4,8,8,2,0,4,6,6,5,
2,1,3,8,4,1,4,6,9,5,1,9,4,1,5,1,1,6,0,9,4,3,3,0,5,7,2,7,0,3,6,5,7,5,9,5,9,1,
9,5,3,0,9,2,1,8,6,1,1,7,3,8,1,9,3,2,6,1,1,7,9,3,1,0,5,1,1,8,5,4,8,0,7,4,4,6,
2,3,7,9,9,6,2,7,4,9,5,6,7,3,5,1,8,8,5,7,5,2,7,2,4,8,9,1,2,2,7,9,3,8,1,8,3,0,
1,1,9,4,9,1);
xeu: array[1..500] of byte =
(2,7,1,8,2,8,1,8,2,8,4,5,9,0,4,5,2,3,5,3,6,0,2,8,7,4,7,1,3,5,2,6,6,2,4,9,7,7,
5,7,2,4,7,0,9,3,6,9,9,9,5,9,5,7,4,9,6,6,9,6,7,6,2,7,7,2,4,0,7,6,6,3,0,3,5,3,
5,4,7,5,9,4,5,7,1,3,8,2,1,7,8,5,2,5,1,6,6,4,2,7,4,2,7,4,6,6,3,9,1,9,3,2,0,0,
3,0,5,9,9,2,1,8,1,7,4,1,3,5,9,6,6,2,9,0,4,3,5,7,2,9,0,0,3,3,4,2,9,5,2,6,0,5,
9,5,6,3,0,7,3,8,1,3,2,3,2,8,6,2,7,9,4,3,4,9,0,7,6,3,2,3,3,8,2,9,8,8,0,7,5,3,
1,9,5,2,5,1,0,1,9,0,1,1,5,7,3,8,3,4,1,8,7,9,3,0,7,0,2,1,5,4,0,8,9,1,4,9,9,3,
4,8,8,4,1,6,7,5,0,9,2,4,4,7,6,1,4,6,0,6,6,8,0,8,2,2,6,4,8,0,0,1,6,8,4,7,7,4,
1,1,8,5,3,7,4,2,3,4,5,4,4,2,4,3,7,1,0,7,5,3,9,0,7,7,7,4,4,9,9,2,0,6,9,5,5,1,
7,0,2,7,6,1,8,3,8,6,0,6,2,6,1,3,3,1,3,8,4,5,8,3,0,0,0,7,5,2,0,4,4,9,3,3,8,2,
6,5,6,0,2,9,7,6,0,6,7,3,7,1,1,3,2,0,0,7,0,9,3,2,8,7,0,9,1,2,7,4,4,3,7,4,7,0,
4,7,2,3,0,6,9,6,9,7,7,2,0,9,3,1,0,1,4,1,6,9,2,8,3,6,8,1,9,0,2,5,5,1,5,1,0,8,
6,5,7,4,6,3,7,7,2,1,1,1,2,5,2,3,8,9,7,8,4,4,2,5,0,5,6,9,5,3,6,9,6,7,7,0,7,8,
5,4,4,9,9,6,9,9,6,7,9,4,6,8,6,4,4,5,4,9,0,5,9,8,7,9,3,1,6,3,6,8,8,9,2,3,0,0,
9,8,7,9,3,1);
xl10: array[1..500] of byte =
(2,3,0,2,5,8,5,0,9,2,9,9,4,0,4,5,6,8,4,0,1,7,9,9,1,4,5,4,6,8,4,3,6,4,2,0,7,6,
0,1,1,0,1,4,8,8,6,2,8,7,7,2,9,7,6,0,3,3,3,2,7,9,0,0,9,6,7,5,7,2,6,0,9,6,7,7,
3,5,2,4,8,0,2,3,5,9,9,7,2,0,5,0,8,9,5,9,8,2,9,8,3,4,1,9,6,7,7,8,4,0,4,2,2,8,
6,2,4,8,6,3,3,4,0,9,5,2,5,4,6,5,0,8,2,8,0,6,7,5,6,6,6,6,2,8,7,3,6,9,0,9,8,7,
8,1,6,8,9,4,8,2,9,0,7,2,0,8,3,2,5,5,5,4,6,8,0,8,4,3,7,9,9,8,9,4,8,2,6,2,3,3,
1,9,8,5,2,8,3,9,3,5,0,5,3,0,8,9,6,5,3,7,7,7,3,2,6,2,8,8,4,6,1,6,3,3,6,6,2,2,
2,2,8,7,6,9,8,2,1,9,8,8,6,7,4,6,5,4,3,6,6,7,4,7,4,4,0,4,2,4,3,2,7,4,3,6,5,1,
5,5,0,4,8,9,3,4,3,1,4,9,3,9,3,9,1,4,7,9,6,1,9,4,0,4,4,0,0,2,2,2,1,0,5,1,0,1,
7,1,4,1,7,4,8,0,0,3,6,8,8,0,8,4,0,1,2,6,4,7,0,8,0,6,8,5,5,6,7,7,4,3,2,1,6,2,
2,8,3,5,5,2,2,0,1,1,4,8,0,4,6,6,3,7,1,5,6,5,9,1,2,1,3,7,3,4,5,0,7,4,7,8,5,6,
9,4,7,6,8,3,4,6,3,6,1,6,7,9,2,1,0,1,8,0,6,4,4,5,0,7,0,6,4,8,0,0,0,2,7,7,5,0,
2,6,8,4,9,1,6,7,4,6,5,5,0,5,8,6,8,5,6,9,3,5,6,7,3,4,2,0,6,7,0,5,8,1,1,3,6,4,
2,9,2,2,4,5,5,4,4,0,5,7,5,8,9,2,5,7,2,4,2,0,8,2,4,1,3,1,4,6,9,5,6,8,9,0,1,6,
7,5,8,9,4,0);

TYPE

pbuf = ^bufnum;

VAR

xstatus : byte;
usedig : word;
rc,c2,c3,c4,c5,c6,c7 : mp;
b1,b2,b3,b4,b5 : bufnum;
ub1,ub2,ub3 : array[1..(maxbuf div 8)+1] of comp;
mpc,re1 : comp;
rr : pbuf;
rp : array[1..maxbuf] of pbuf;
ac : array[0..20] of byte;
s,s1 : string;
r1,d,t,b,z : byte;
r : shortint;
m,lv,lr,nc,i,j,x,ia3,da3,ncprd,ndprd,niprd,y,k1,dx,k2,dy,ncqur,ndqur,
m1,m2,ncvr,ndvr,ncfr,ndfr,ncrq,ncmq,ncpq,n,c,hq,n1,n2,n3,zp,np,kf,sps : word;
h,md1,md2,rf,uf,df,ef : longint;
sgdr1,sgdr2,sss,sg1,sg2,sg3,powsign,subflag : boolean;
p,ps,ip,zs,ls,nd : integer;
q : longint;

{* internal procedures *}

procedure xelz (var a: mp);                 {eliminates the superfluous zeroes}
begin
if (a.c[a.n]=0) and (a.d>0) then begin
while (a.c[a.n]=0) and (a.n>a.i) do dec(a.n);
a.d:=a.n-a.i;
end;
j:=1;
while (a.c[j]=0) and (j<a.i) do inc(j);
if j>1 then begin
dec(j);
dec(a.n,j); dec(a.i,j);
for i:=1 to a.n do a.c[i]:=a.c[i+j];
end
end;

procedure recip;                        {calculates in rc the reciprocal of uf}
begin
kf:=1; rf:=10;
while uf>rf do begin
rf:=rf*10; inc(kf); rc.c[kf]:=0;
end;
while (kf<usedig) and (rf>0) do begin
inc(kf); rc.c[kf]:=rf div uf; rf:=(rf mod uf)*10;
end;
rc.n:=kf; rc.d:=kf-1;
end;

procedure quot;                           {calculates in rc the quotient ef/uf}
begin
kf:=1; rf:=ef*10;
while uf>rf do begin
rf:=rf*10; inc(kf); rc.c[kf]:=0;
end;
while (kf<usedig) and (rf>0) do begin
inc(kf); rc.c[kf]:=rf div uf; rf:=(rf mod uf)*10;
end;
rc.n:=kf; rc.d:=kf-1;
end;

procedure xtox (a: mp; var b: mp);                           {a is copied in b}
begin
b.n:=a.n; b.i:=a.i; b.d:=a.d; b.s:=a.s;
move(a.c,b.c,a.n);
end;

procedure xto0 (var a: mp);                                     {stores 0 on a}
begin
a.c[1]:=0; a.n:=1; a.i:=1; a.d:=0; a.s:=true;
end;

procedure xto1 (var a: mp);                                     {stores 1 on a}
begin
a.c[1]:=1; a.n:=1; a.i:=1; a.d:=0; a.s:=true;
end;

procedure xtom1 (var a: mp);                                   {stores -1 on a}
begin
a.c[1]:=1; a.n:=1; a.i:=1; a.d:=0; a.s:=false;
end;

procedure xto2 (var a: mp);                                     {stores 2 on a}
begin
a.c[1]:=2; a.n:=1; a.i:=1; a.d:=0; a.s:=true;
end;

function xeq(a,b: mp): boolean;      {comparison: true if a=b, false otherwise}
begin
if (a.n<>b.n) or (a.i<>b.i) or (a.s<>b.s) then xeq:=false
else begin
i:=1;
while (a.c[i]=b.c[i]) and (i<a.n) do inc(i);
xeq:=a.c[i]=b.c[i];
end;
end;

function xgt(a,b: mp): boolean;      {comparison: true if a>b, false otherwise}
begin
if a.s<>b.s then xgt:=a.s
else begin
if a.i<>b.i then begin
if a.s then xgt:=a.i>b.i else xgt:=a.i<b.i
end
else begin
i:=1;
while (a.c[i]=b.c[i]) and (i<a.i) do inc(i);
if a.c[i]<>b.c[i] then begin
if a.s then xgt:=a.c[i]>b.c[i] else xgt:=a.c[i]<b.c[i];
end
else begin
if (a.d=0) and (b.d=0) then xgt:=false
else begin
if (a.d=0) or (b.d=0) then begin
if a.s then xgt:=a.d>0 else xgt:=a.d=0;
end
else begin
i:=a.i+1;
j:=1;
while (j<a.d) and (j<b.d) and (a.c[i]=b.c[i]) do begin
inc(i);
inc(j);
end;
if a.c[i]<>b.c[i] then begin
if a.s then xgt:=a.c[i]>b.c[i] else xgt:=a.c[i]<b.c[i];
end
else begin
if a.s then xgt:=a.d>b.d else xgt:=a.d<b.d;
end;
end;
end;
end;
end;
end;
end;

function xfrome (e: ex): mp;                       {converts an extended to mp}
begin
xstatus:=0;
str(e,s);
b:=pos('E',s);
x:=b-3;
val(copy(s,b+1,length(s)-b),nd,ip);
delete(s,3,1);
xfrome.s:=s[1]=' ';
delete(s,1,1);
if nd=0 then begin
for i:=1 to x do xfrome.c[i]:=ord(s[i])-48;
xfrome.n:=x; xfrome.i:=1; xfrome.d:=x-1;
end
else if nd<0 then begin
if nd>=usedig then begin
xto0(xfrome);
exit;
end;
nd:=-nd;
fillchar(xfrome.c,nd,0);
for i:=1 to x do xfrome.c[nd+i]:=ord(s[i])-48;
xfrome.n:=x+nd; xfrome.i:=1; xfrome.d:=xfrome.n-1;
end
else if nd<x then begin
for i:=1 to x do xfrome.c[i]:=ord(s[i])-48;
xfrome.n:=x; xfrome.i:=nd+1; xfrome.d:=xfrome.n-xfrome.i;
end
else begin
if nd+1>usedig then begin
xto0(xfrome);
xstatus:=1;
exit;
end;
for i:=1 to x do xfrome.c[i]:=ord(s[i])-48;
for i:=1 to nd-x+1 do xfrome.c[x+i]:=0;
xfrome.n:=nd+1; xfrome.i:=xfrome.n; xfrome.d:=0;
end;
if xfrome.n>usedig then
if xfrome.i>usedig then begin
xto0(xfrome);
xstatus:=1;
exit
end
else begin
xfrome.n:=usedig;
xfrome.d:=usedig-xfrome.i;
end;
xelz(xfrome);
end;

function efromx (a: mp): ex;                       {converts an mp to extended}
begin
xstatus:=0;
c2:=a;
xelz(c2);
if (c2.c[1]=0) and (c2.n=1) then begin
efromx:=0;
exit;
end;
i:=1;
while (c2.c[i]=0) and (i<c2.n) do inc(i);
dec(i);
if i=0 then nd:=a.i-1 else nd:=-i;
j:=c2.n-i;
if j>18 then y:=19 else y:=j;
if i>0 then for j:=1 to y do c2.c[j]:=c2.c[j+i];
if y>18 then begin
if c2.c[y]>5 then
if c2.c[y-1]<9 then inc(c2.c[y-1])
else begin
j:=y-1;
while (c2.c[j]=9) and (j>0) do begin
c2.c[j]:=0;
dec(j);
end;
if j>0 then inc(c2.c[j])
else begin
inc(nd);
c2.c[1]:=1;
end;
end;
dec(y);
end
else if y<18 then for j:=y+1 to 18 do c2.c[j]:=0;
s:='                  ';
for j:=1 to 18 do s[j]:=chr(c2.c[j]+48);
if a.s then s:=' '+s else s:='-'+s;
insert('.',s,3);
str(abs(nd),s1);
while length(s1)<4 do s1:='0'+s1;
if nd<0 then s1:='E-'+s1 else s1:='E+'+s1;
s:=s+s1;
val(s,efromx,ip);
end;

{ addition: a3:=a1+a2 }

procedure xadd (a1,a2: mp; var a3: mp);
begin
xstatus:=0;
n3:=a1.n;
n2:=a2.n;
sg1:=a1.s;
if subflag then sg2:=(not a2.s) else sg2:=a2.s;
x:=a1.i;
y:=a2.i;
if x>y then begin
move(a1.c,b1,n3);
dec(x,y);
for i:=1 to n2 do b2[i+x]:=a2.c[i];
fillchar(b2,x,0);
inc(n2,x);
end
else if x<y then begin
move(a2.c,b2,n2);
dec(y,x);
for i:=1 to n3 do b1[i+y]:=a1.c[i];
fillchar(b1,y,0);
inc(n3,y);
end
else begin
move(a1.c,b1,n3);
move(a2.c,b2,n2);
end;
x:=a1.d;
y:=a2.d;
if x>y then begin
da3:=x;
dec(x,y);
for i:=1 to x do b2[n2+i]:=0;
end
else if x<y then begin
da3:=y;
dec(y,x);
for i:=1 to y do b1[n3+i]:=0;
inc(n3,y);
end
else da3:=x;
if sg1=sg2 then begin
sg3:=sg1;
z:=0;
for i:=n3 downto 1 do begin
z:=b1[i]+b2[i]+ck[z];
b3[i]:=m10[z];
end;
if z>9 then begin
for i:=n3 downto 1 do b3[i+1]:=b3[i];
b3[1]:=1;
inc(n3)
end;
end
else begin
i:=1;
while (b1[i]=b2[i]) and (i<n3) do inc(i);
if b1[i]=b2[i] then begin
xto0(a3);
exit;
end;
if b1[i]>b2[i] then begin
sg3:=sg1;
r:=0;
for i:=n3 downto 1 do begin
r:=b1[i]-b2[i]-cu[r];
b3[i]:=cd[r];
end;
end
else begin
sg3:=sg2;
r:=0;
for i:=n3 downto 1 do begin
r:=b2[i]-b1[i]-cu[r];
b3[i]:=cd[r];
end;
end;
end;
ia3:=n3-da3;
if (ia3>1) and (b3[1]=0) then begin
i:=1;
while (i<ia3) and (b3[i]=0) do inc(i);
if i>1 then begin
dec(i);
dec(n3,i);
dec(ia3,i);
for j:=1 to n3 do b3[j]:=b3[j+i];
end;
end;
if n3>usedig then
if ia3>usedig then xstatus:=5
else begin
n3:=usedig;
da3:=n3-ia3;
end;
if (b3[n3]=0) and (da3>0) then begin
while (b3[n3]=0) and (n3>ia3) do dec(n3);
da3:=n3-ia3;
end;
move(b3,a3.c,n3);
a3.n:=n3; a3.d:=da3; a3.i:=ia3; a3.s:=sg3;
end;

{ subtraction: a3:=a1-a2 } 

procedure xsub (a1,a2: mp; var a3: mp);
begin
subflag:=true;
xadd(a1,a2,a3);
subflag:=false;
if xstatus>0 then xstatus:=6;
end;

{ multiplication: a3:=a1*a2. The routine divides the digits in groups of
eight, uses comp variables to perform a simplified polynomial calculus
of the product, then converts the comp values of the result to single digits }

procedure xmul (a1,a2: mp; var a3: mp);
begin
n1:=(a1.n div 8)+1;
m:=2;
j:=(a1.n mod 8)+1;
case j of
8: ub1[1]:=a1.c[1]*p6+a1.c[2]*p5+a1.c[3]*p4+a1.c[4]*p3+a1.c[5]*p2+a1.c[6]*p1
+a1.c[7];
7: ub1[1]:=a1.c[1]*p5+a1.c[2]*p4+a1.c[3]*p3+a1.c[4]*p2+a1.c[5]*p1+a1.c[6];
6: ub1[1]:=a1.c[1]*p4+a1.c[2]*p3+a1.c[3]*p2+a1.c[4]*p1+a1.c[5];
5: ub1[1]:=a1.c[1]*p3+a1.c[2]*p2+a1.c[3]*p1+a1.c[4];
4: ub1[1]:=a1.c[1]*p2+a1.c[2]*p1+a1.c[3];
3: ub1[1]:=a1.c[1]*p1+a1.c[2];
2: ub1[1]:=a1.c[1];
else begin
dec(m);
dec(n1);
end;
end;
for x:=m to n1 do begin
ub1[x]:=a1.c[j]*p7+a1.c[j+1]*p6+a1.c[j+2]*p5+a1.c[j+3]*p4+a1.c[j+4]*p3
+a1.c[j+5]*p2+a1.c[j+6]*p1+a1.c[j+7];
inc(j,8);
end;
n2:=(a2.n div 8)+1;
m:=2;
j:=(a2.n mod 8)+1;
case j of
8: ub2[1]:=a2.c[1]*p6+a2.c[2]*p5+a2.c[3]*p4+a2.c[4]*p3+a2.c[5]*p2+a2.c[6]*p1
+a2.c[7];
7: ub2[1]:=a2.c[1]*p5+a2.c[2]*p4+a2.c[3]*p3+a2.c[4]*p2+a2.c[5]*p1+a2.c[6];
6: ub2[1]:=a2.c[1]*p4+a2.c[2]*p3+a2.c[3]*p2+a2.c[4]*p1+a2.c[5];
5: ub2[1]:=a2.c[1]*p3+a2.c[2]*p2+a2.c[3]*p1+a2.c[4];
4: ub2[1]:=a2.c[1]*p2+a2.c[2]*p1+a2.c[3];
3: ub2[1]:=a2.c[1]*p1+a2.c[2];
2: ub2[1]:=a2.c[1];
else begin
dec(m);
dec(n2);
end;
end;
for x:=m to n2 do begin
ub2[x]:=a2.c[j]*p7+a2.c[j+1]*p6+a2.c[j+2]*p5+a2.c[j+3]*p4+a2.c[j+4]*p3
+a2.c[j+5]*p2+a2.c[j+6]*p1+a2.c[j+7];
inc(j,8);
end;
n3:=n1+n2;
m:=n1+1;
x:=n2+1;
n:=2;
mpc:=0;
for i:=n3 downto 2 do begin
if i > x then nc:=i-n2 else nc:=1;
if i > m then lv:=n1 else lv:=i-1;
for j:=nc to lv do mpc:=mpc+ub1[j]*ub2[i-j];
re1:=int(mpc/vcc);
ub3[i]:=mpc-re1*vcc;
mpc:=re1;
end;
if mpc=0 then while (ub3[n]=0) and (n<n3) do inc(n)
else begin
ub3[1]:=mpc;
dec(n);
end;
xstatus:=0;
q:=trunc(ub3[n]);
j:=0;
repeat
ac[j]:=q mod 10;
q:=q div 10;
inc(j);
until q=0;
ncprd:=j+((n3-n) shl 3);
ndprd:=a1.d+a2.d;
if ndprd<ncprd then begin
niprd:=ncprd-ndprd;
ncprd:=0;
if niprd>usedig then begin
xstatus:=7;
xto0(a3);
exit;
end
end
else begin
ncprd:=ndprd-ncprd+1;
if ncprd<usedig then begin
fillchar(a3.c,ncprd,0);
niprd:=1;
end
else begin
xto0(a3);
exit;
end;
end;
for i:=1 to j do a3.c[ncprd+i]:=ac[j-i];
inc(ncprd,j);
while (n<n3) and (ncprd<usedig) do begin
inc(n);
q:=trunc(ub3[n]);
inc(ncprd);
a3.c[ncprd]:=q div p7;
q:=q mod p7;
inc(ncprd);
a3.c[ncprd]:=q div p6;
q:=q mod p6;
inc(ncprd);
a3.c[ncprd]:=q div p5;
q:=q mod p5;
inc(ncprd);
a3.c[ncprd]:=q div p4;
q:=q mod p4;
inc(ncprd);
a3.c[ncprd]:=q div p3;
q:=q mod p3;
inc(ncprd);
a3.c[ncprd]:=q div p2;
q:=q mod p2;
inc(ncprd);
a3.c[ncprd]:=q div p1;
inc(ncprd);
a3.c[ncprd]:=q mod p1;
end;
if ncprd>usedig then ncprd:=usedig;
ndprd:=ncprd-niprd;
if (a3.c[ncprd]=0) and (ndprd>0) then begin
while (a3.c[ncprd]=0) and (ncprd>niprd) do dec(ncprd);
ndprd:=ncprd-niprd;
end;
a3.n:=ncprd; a3.d:=ndprd; a3.i:=niprd;
if (a3.n=1) and (a3.c[1]=0) then a3.s:=true else a3.s:=(a1.s=a2.s);
end;

{ division: a3:=a1/a2. The routine operates normally with progressive
partial quotients, using a sliding pointer to change position of the
remainder in buffer without move the elements }

procedure xdiv (a1,a2: mp; var a3: mp);
label cycle1,cycle2,quit;
begin
xstatus:=0;
k1:=a1.n; dx:=a1.d; sgdr1:=a1.s;
move(a1.c,b1,k1);
k2:=a2.n; dy:=a2.d; sgdr2:=a2.s;
move(a2.c,b2,k2);
i:=k1-dx;
if b1[k1]=0 then begin
while (b1[k1]=0) and (k1>i) do dec(k1);
dx:=k1-i;
end;
j:=1;
while (b1[j]=0) and (j<i) do inc(j);
if j>1 then begin
dec(j);
dec(k1,j);
for i:=1 to k1 do b1[i]:=b1[i+j];
end;
i:=k2-dy;
if b2[k2]=0 then begin
while (b2[k2]=0) and (k2>i) do dec(k2);
dy:=k2-i;
end;
j:=1;
while (b2[j]=0) and (j<i) do inc(j);
if j>1 then begin
dec(j);
dec(k2,j);
for i:=1 to k2 do b2[i]:=b2[i+j];
end;
if (b2[1]=0) and (k2=1) then begin
xto0(a3);
xstatus:=9;
exit;
end;
if (b1[1]=0) and (k1=1) then begin
xto0(a3);
exit;
end;
i:=1;
while (b1[i]=0) and (i<k1) do inc(i);
if i>1 then begin
dec(i);
dec(k1,i);
dec(dx,i);
for j:=1 to k1 do b1[j]:=b1[j+i];
for j:=1 to i do b2[k2+j]:=0;
inc(k2,i);
end;
i:=1;
while (b2[i]=0) and (i<k2) do inc(i);
if i>1 then begin
dec(i);
dec(k2,i);
dec(dy,i);
for j:=1 to k2 do b2[j]:=b2[j+i];
for j:=1 to i do b1[k1+j]:=0;
inc(k1,i);
end;
if dx<dy then begin
i:=dy-dx;
for j:=1 to i do b1[k1+j]:=0;
inc(k1,i);
end
else if dy<dx then begin
i:=dx-dy;
for j:=1 to i do b2[k2+j]:=0;
inc(k2,i);
end;
ndqur:=0;
ncqur:=0;
if k1<k2 then begin
j:=k2-k1;
for i:=k1+1 to k2 do b1[i]:=0;
fillchar(b3,j,0);
ndqur:=j;
ncqur:=j;
k1:=k2;
end;
if k1=k2 then begin
i:=1;
while (b1[i]=b2[i]) and (i<k1) do inc(i);
if b1[i]<b2[i] then begin
inc(k1);
b1[k1]:=0;
inc(ncqur);
b3[ncqur]:=0;
inc(ndqur);
end;
end;
sps:=1;
rr:=rp[1];
i:=1;
while (b1[i]=b2[i]) and (i<k2) do inc(i);
if b1[i]<b2[i] then nc:=k2+1 else nc:=k2;
move(b1,rr^,nc);
lr:=nc;
if k2<5 then b:=k2 else b:=5;
case b of
1 : md2:=b2[1];
2 : md2:=b2[1]*p1+b2[2];
3 : md2:=b2[1]*p2+b2[2]*p1+b2[3];
4 : md2:=b2[1]*p3+b2[2]*p2+b2[3]*p1+b2[4];
5 : md2:=b2[1]*p4+b2[2]*p3+b2[3]*p2+b2[4]*p1+b2[5]+1;
end;
cycle1:
if lr>k2 then b:=6 else b:=5;
if lr<b then b:=lr;
case b of
1 : md1:=rr^[1];
2 : md1:=rr^[1]*p1+rr^[2];
3 : md1:=rr^[1]*p2+rr^[2]*p1+rr^[3];
4 : md1:=rr^[1]*p3+rr^[2]*p2+rr^[3]*p1+rr^[4];
5 : md1:=rr^[1]*p4+rr^[2]*p3+rr^[3]*p2+rr^[4]*p1+rr^[5];
6 : md1:=rr^[1]*p5+rr^[2]*p4+rr^[3]*p3+rr^[4]*p2+rr^[5]*p1+rr^[6];
end;
h:=md1 div md2;
if lr>k2 then begin
inc(sps);
rr:=rp[sps];
dec(lr);
end;
r1:=0;
r:=0;
for i:=lr downto 1 do begin
r1:=b2[i]*h+d10[r1];
r:=rr^[i]-m10[r1]-cu[r];
rr^[i]:=cd[r];
end;
while (rr^[i]=b2[i]) and (i<lr) do inc(i);
while rr^[i]>=b2[i] do begin
inc(h);
r:=0;
for i:=lr downto 1 do begin
r:=rr^[i]-b2[i]-cu[r];
rr^[i]:=cd[r];
end;
while (rr^[i]=b2[i]) and (i<lr) do inc(i);
end;
cycle2:
inc(ncqur);
b3[ncqur]:=h;
j:=1;
while (rr^[j]=0) and (j<lr) do inc(j);
if j>1 then begin
dec(j);
dec(lr,j);
inc(sps,j);
rr:=rp[sps];
end;
if (ncqur>=usedig) or ((lr=1) and (rr^[1]=0) and (ndqur>0)) then goto quit;
inc(nc);
inc(lr);
if (nc>k1) then begin
rr^[lr]:=0;
inc(ndqur);
end
else rr^[lr]:=b1[nc];
if lr>k2 then goto cycle1;
if lr=k2 then begin
i:=1;
while (rr^[i]=b2[i]) and (i<k2) do inc(i);
if (rr^[i]<b2[i]) then h:=0
else goto cycle1;
end
else h:=0;
goto cycle2;
quit:
i:=ncqur-ndqur;
j:=1;
while (b3[j]=0) and (j<i) do inc(j);
if j>1 then begin
dec(j);
dec(ncqur,j); dec(i,j);
for x:=1 to ncqur do b3[x]:=b3[x+j];
end;
if ncqur>usedig then
if i>usedig then xstatus:=8
else begin
ncqur:=usedig;
ndqur:=ncqur-i;
end;
if (b3[ncqur]=0) and (ndqur>0) then begin
while (b3[ncqur]=0) and (ncqur>i) do dec(ncqur);
ndqur:=ncqur-i;
end;
move(b3,a3.c,ncqur);
a3.n:=ncqur; a3.d:=ndqur; a3.i:=i; a3.s:=a1.s=a2.s;
end;

{ the square root uses the normal method to achieve a reliable precision }

procedure xsqrt (a1: mp; var a2: mp);
label ref1,ref2,ref3;
begin
xstatus:=0;
if a1.s=false then begin
xstatus:=10;
xto0(a2);
exit;
end;
move(a1.c,b2,a1.n);
ndfr:=a1.d;
ncfr:=a1.n;
i:=1;
while (b2[i]=0) and (i<ncfr) do inc(i);
if i>1 then begin
dec(i);
dec(ncfr,i);
for j:=1 to ncfr do b2[j]:=b2[j+i];
end;
if odd(ndfr) then begin
inc(ndfr);
inc(ncfr);
b2[ncfr]:=0;
end;
n:=0;
c:=2-(ncfr and 1);
if c=1 then p:=b2[1] else p:=b2[1]*10+b2[2];
hq:=tr[p];
m:=p-hq*hq;
b1[1]:=hq;
ncvr:=1;
if m<10 then begin
b5[1]:=m;
ncrq:=1
end
else begin
b5[1]:=d10[m];
b5[2]:=m10[m];
ncrq:=2;
end;
while n<usedig do begin
inc(n);
if c<ncfr then begin
b5[ncrq+1]:=b2[c+1];
b5[ncrq+2]:=b2[c+2];
end
else begin
b5[ncrq+1]:=0;
b5[ncrq+2]:=0;
end;
inc(ncrq,2);
inc(c,2);
x:=1;
while (b5[x]=0) and (x<ncrq) do inc(x);
if x>1 then begin
dec(x);
dec(ncrq,x);
for m:=1 to ncrq do b5[m]:=b5[m+x];
end;
if b5[1]=0 then begin
inc(ncvr);
b1[ncvr]:=0;
goto ref2;
end;
ncmq:=1;
z:=0;
for x:=ncvr downto 1 do begin
z:=b1[x]*2+d10[z];
inc(ncmq);
b3[ncmq]:=m10[z];
end;
if z>9 then begin
inc(ncmq);
b3[ncmq]:=d10[z];
end;
if ncmq<5 then p:=2 else p:=ncmq-2;
d:=ncmq-p;
m2:=b3[p];
j:=0;
while j<d do begin
inc(j);
inc(m2,b3[p+j]*dpot[j]);
end;
inc(d);
t:=d+1;
if ncrq<t then z:=ncrq else z:=t;
b:=z;
m1:=b5[b];
while (b>1) do begin
dec(b);
inc(m1,b5[b]*dpot[z-b]);
end;
if z>d then begin
j:=m1 div dpot[z-d];
if j>m2 then m1:=j;
end;
hq:=(m1+1) div m2;
if hq=0 then begin
inc(ncvr);
b1[ncvr]:=0;
goto ref1;
end;
if hq>9 then hq:=9;
ref3:
b3[1]:=hq;
z:=0;
i:=ncmq+1;
for x:=1 to ncmq do begin
z:=b3[x]*hq+d10[z];
dec(i);
b4[i]:=m10[z];
end;
ncpq:=ncmq;
if z>9 then begin
for x:=ncpq downto 1 do b4[x+1]:=b4[x];
b4[1]:=d10[z];
inc(ncpq);
end;
if ncpq>ncrq then begin
dec(hq);
if hq>0 then goto ref3
else begin
inc(ncvr);
b1[ncvr]:=0;
goto ref1;
end
end;
if ncpq=ncrq then begin
x:=1;
while (b4[x]=b5[x]) and (x<ncpq) do inc(x);
if b4[x]>b5[x] then begin
dec(hq);
if hq>0 then goto ref3
else begin
inc(ncvr);
b1[ncvr]:=0;
goto ref1;
end
end;
if b4[x]=b5[x] then begin
ncrq:=1;
b5[1]:=0;
inc(ncvr);
b1[ncvr]:=hq;
goto ref2;
end
end;
inc(ncvr);
b1[ncvr]:=hq;
if ncrq>ncpq then begin
j:=ncrq-ncpq;
for i:=ncpq downto 1 do b4[i+j]:=b4[i];
fillchar(b4,j,0);
end;
r:=0;
for x:=ncrq downto 1 do begin
r:=b5[x]-b4[x]-cu[r];
b5[x]:=cd[r];
end;
ref1:
x:=1;
while (b5[x]=0) and (x<ncrq) do inc(x);
if x>1 then begin
dec(x);
dec(ncrq,x);
for m:=1 to ncrq do b5[m]:=b5[m+x];
end;
ref2:
end;
i:=a1.i;
if i>1 then begin
i:=(i+1) div 2;
ndvr:=ncvr-i;
end
else begin
if a1.d>ncfr then begin
i:=((ndfr-ncfr) div 2)+1;
for j:=ncvr downto 1 do b1[j+i]:=b1[j];
fillchar(b1,i,0);
inc(ncvr,i);
ndvr:=ncvr-1;
end
else begin
if a1.c[1]=0 then begin
for j:=ncvr downto 1 do b1[j+1]:=b1[j];
b1[1]:=0;
inc(ncvr);
end;
ndvr:=ncvr-1;
end
end;
j:=ncvr-ndvr;
if (j>1) and (b1[1]=0) then begin
i:=1;
while (i<j) and (b1[i]=0) do inc(i);
if i>1 then begin
dec(i);
dec(ncvr,i);
dec(j,i);
for x:=1 to ncvr do b1[x]:=b1[x+i];
end;
end;
if (ncvr>usedig) and (j<=usedig) then begin
ncvr:=usedig;
ndvr:=ncvr-j;
end;
if (b1[ncvr]=0) and (ndvr>0) then begin
while (b1[ncvr]=0) and (ncvr>j) do dec(ncvr);
ndvr:=ncvr-j;
end;
move(b1,a2.c,ncvr);
a2.n:=ncvr; a2.d:=ndvr; a2.i:=j; a2.s:=true;
end;

{ the exponential uses the classic series:
exp(x) = 1 + x/1! + x**2/2! + x**3/3! + ... with reductions }

procedure xexp (a1: mp; var a2: mp);
begin
xstatus:=0;
xtox(a1,c4);
xelz(c4);
if c4.n=1 then
if c4.c[1]=0 then begin
xto1(a2);
exit;
end
else if c4.c[1]=1 then begin
move(xeu,a2.c,usedig);
a2.i:=1; a2.d:=usedig-1; a2.n:=usedig; a2.s:=true;
if (not c4.s) then begin
xto1(c4);
xdiv(c4,a2,a2);
end;
exit;
end;
c4.s:=true;
if c4.c[1]>=1 then begin
zp:=c4.i-1;
c4.i:=1;
c4.d:=c4.d+zp;
end
else zp:=0;
xto1(c5);
xto1(c6);
uf:=0;
repeat
inc(uf);
recip;
xmul(c6,c4,c6);
if xstatus=0 then xmul(c6,rc,c6);
xadd(c5,c6,c5);
until ((c6.n=1) and (c6.c[1]=0)) or (xstatus>0);
if xstatus>0 then begin
xto0(a2);
if a1.s then xstatus:=19 else xstatus:=0;
exit;
end;
if zp>0 then begin
inc(zp);
c4.c[1]:=1;
for i:=2 to zp do c4.c[i]:=0;
c4.n:=zp;
c4.d:=0;
c4.i:=zp;
c4.s:=true;
xto1(c6);
xto2(c3);
while c4.s and ((c4.c[1]>0) or (c4.n>1)) do begin
sss:=(c4.c[c4.n] and 1>0);
xdiv(c4,c3,c4);
c4.n:=c4.i; c4.d:=0;
if sss then begin
xmul(c6,c5,c6);
if xstatus>0 then begin
if a1.s then xstatus:=19 else xstatus:=0;
xto0(a2);
exit;
end;
end;
if (c4.c[1]>0) or (c4.n>1) then begin
xmul(c5,c5,c5);
if xstatus>0 then begin
if a1.s then xstatus:=19 else xstatus:=0;
xto0(a2);
exit;
end;
end;
end;
xtox(c6,c5);
end;
if (not a1.s) then begin
xto1(c3);
xdiv(c3,c5,a2);
end
else xtox(c5,a2);
end;

{ the logarithm uses the classic series:
ln(x) = 2 * (z + z**3/3 + z**5/5 + ... ) where z=(x-1)/(x+1),
with reductions }

procedure xln (a1: mp; var a2: mp);
begin
xto0(c7);
if xgt(a1,c7) then xstatus:=0
else begin
xstatus:=20;
xtox(c7,a2);
exit;
end;
xtox(a1,c4);
xelz(c4);
if (c4.c[1]=1) and (c4.n=1) then begin
xto0(a2);
exit;
end;
if c4.c[1]=0 then begin
j:=1;
i:=c4.n;
while (c4.c[j]=0) and (j<i) do inc(j);
dec(j);
dec(c4.n,j); c4.i:=1; c4.d:=c4.n-1;
for i:=1 to a1.n do c4.c[i]:=c4.c[i+j];
np:=j;
sss:=false;
end
else begin
sss:=true;
np:=c4.i-1;
c4.i:=1;
c4.d:=c4.d+np;
end;
inc(usedig);
xsqrt(c4,c4);
xsqrt(c4,c4);
xsqrt(c4,c4);
c7.c[1]:=1;
xadd(c4,c7,c5);
c7.s:=false;
xadd(c4,c7,c6);
xdiv(c6,c5,c4);
xtox(c4,c5);
uf:=1;
xmul(c4,c4,c6);
repeat
inc(uf,2);
recip;
xmul(c4,c6,c4);
xmul(c4,rc,c3);
xadd(c5,c3,c5);
until (c3.n=1) and (c3.c[1]=0);
dec(usedig);
c7.c[1]:=1; c7.c[2]:=6;
c7.n:=2; c7.i:=2; c7.s:=true;
xmul(c5,c7,c5);
if np>0 then begin
move(xl10,c7.c,usedig);
c7.i:=1; c7.d:=usedig-1; c7.n:=usedig; c7.s:=true;
j:=0;
repeat
ac[j]:=np mod 10;
np:=np div 10;
inc(j);
until np=0;
for i:=1 to j do c4.c[i]:=ac[j-i];
c4.n:=j; c4.i:=j; c4.d:=0; c4.s:=true;
xmul(c7,c4,c7);
if sss then xadd(c5,c7,c5) else xsub(c5,c7,c5);
end;
c5.s:=sss;
xtox(c5,a2);
end;

{ log(x) = ln(x)/ln(10) }

procedure xlog (a1: mp; var a2: mp);
begin
xln(a1,a2);
if xstatus>0 then xstatus:=21
else begin
move(xl10,c3.c,usedig);
c3.i:=1; c3.d:=usedig-1; c3.n:=usedig; c3.s:=true;
xdiv(a2,c3,a2);
end;
end;

{ integer part }

procedure xint (a1: mp; var a2: mp);
begin
a2.n:=a1.i;
a2.d:=0;
a2.i:=a1.i;
a2.s:=a1.s;
move(a1.c,a2.c,a1.i);
if (a2.n=1) and (a2.c[1]=0) then a2.s:=true;
end;

{ fractional part }

procedure xfrac (a1: mp; var a2: mp);
begin
j:=a1.i;
a2.n:=a1.d+1;
a2.i:=1;
a2.d:=a1.d;
a2.s:=a1.s;
a2.c[1]:=0;
for i:=1 to a1.d do a2.c[i+1]:=a1.c[j+i];
if a2.n=1 then a2.s:=true;
end;

{ sign function }

procedure xsign (a1: mp; var a2: mp);
begin
move(a1.c,a2.c,a1.n);
a2.n:=a1.n; a2.i:=a1.i; a2.d:=a1.d; a2.s:=a1.s;
xelz(a2);
if (a2.c[1]>0) or (a2.n>1) then begin
a2.n:=1; a2.i:=1; a2.d:=0; a2.c[1]:=1;
end;
end;

{ ceil function }

procedure xceil (a1: mp; var a2: mp);
begin
move(a1.c,a2.c,a1.n);
a2.n:=a1.n; a2.i:=a1.i; a2.d:=a1.d; a2.s:=a1.s;
xelz(a2);
if (a2.s) and (a2.d>0) then begin
xto1(c2);
xadd(a2,c2,a2);
end;
a2.n:=a2.i; a2.d:=0;
end;

{ floor function }

procedure xfloor(a1: mp; var a2: mp);
begin
move(a1.c,a2.c,a1.n);
a2.n:=a1.n; a2.i:=a1.i; a2.d:=a1.d; a2.s:=a1.s;
xelz(a2);
if (not a2.s) and (a2.d>0) then begin
xtom1(c2);
xadd(a2,c2,a2);
end;
a2.n:=a2.i; a2.d:=0;
end;

{ power: a3:=a1**a2 }

procedure xpow(a1,a2: mp; var a3: mp);
begin
xtox(a1,c3);
xtox(a2,c4);
xelz(c4);
xto0(c5);
if ((xgt(c5,c3) or xeq(c3,c5)) and (c4.d>0))
or (xeq(c4,c5) and (xgt(c5,c3) or xeq(c3,c5)))
or (xeq(c3,c5) and xgt(c5,c4)) then begin
xstatus:=22;
xtox(c5,a3);
exit;
end;
if c4.d>0 then begin
xln(a1,a3);
xmul(a2,a3,a3);
if xstatus=0 then xexp(a3,a3);
if xstatus>0 then begin
xstatus:=23;
xtox(c5,a3);
end;
exit;
end;
c5.c[1]:=1;
powsign:=c4.s;
if (not powsign) then begin
sss:=c3.s;
c3.s:=true;
if xgt(c5,c3) then begin
xdiv(c5,c3,c3);
powsign:=true;
end;
c3.s:=sss;
end;
c4.s:=true;
xto2(c6);
while c4.s and ((c4.c[1]>0) or (c4.n>1)) do begin
sss:=(c4.c[c4.n] and 1>0);
xdiv(c4,c6,c4);
c4.n:=c4.i; c4.d:=0;
if sss then begin
xmul(c5,c3,c5);
if xstatus>0 then begin
xstatus:=23;
c6.c[1]:=0;
xtox(c6,a3);
exit;
end;
end;
if (c4.c[1]>0) or (c4.n>1) then begin
xmul(c3,c3,c3);
if xstatus>0 then begin
xstatus:=23;
c6.c[1]:=0;
xtox(c6,a3);
exit;
end;
end;
end;
if (not powsign) then
if (c5.c[1]=0) and (c5.n=1) then begin
xstatus:=23;
c6.c[1]:=0;
xtox(c6,a3);
exit;
end
else begin
c6.c[1]:=1;
xdiv(c6,c5,c5);
end;
xtox(c5,a3);
end;

{ factorial }

procedure xfact (a1: mp; var a2: mp);
begin
xtox(a1,a2);
xelz(a2);
xto0(c4);
if xgt(c4,a2) or (a2.d>0) then begin
xto0(a2);
xstatus:=33;
exit;
end;
xto1(c3);
xto1(c5);
xstatus:=0;
while xgt(a2,c4) and (xstatus=0) do begin
xadd(c4,c5,c4);
xmul(c3,c4,c3);
end;
if xstatus>0 then begin
xstatus:=34;
xto0(a2);
end
else xtox(c3,a2);
end;

{ the sine uses the classic series:
sin(x) = x - x**3/3! + x**5/5! - x**7/7! + ... with reductions }

procedure xsin (a1: mp; var a2: mp);
begin
xtox(a1,c3);
c3.s:=true;
move(xpg,c4.c,usedig);
c4.i:=1; c4.d:=usedig-1; c4.n:=usedig; c4.s:=true;
xto2(c5);
xmul(c4,c5,c4);
xdiv(c3,c4,c3);
xfrac(c3,c3);
if c3.d>usedig then begin
c3.d:=usedig;
c3.n:=c3.i+usedig;
end;
xelz(c3);
if (c3.c[1]=0) and (c3.n=1) then begin
xto0(a2);
exit;
end;
if (c3.d=2) and (c3.c[2]=7) and (c3.c[3]=5) then begin
xtom1(a2);
exit;
end
else if (c3.c[2]>7) or ((c3.d>1) and (c3.c[2]=7) and (c3.c[3]>=5))
then begin
xto1(c5);
xsub(c5,c3,c3);
sss:=false;
end
else if (c3.d=1) and (c3.c[2]=5) then begin
xto0(a2);
exit;
end
else if c3.c[2]>=5 then begin
c5.c[1]:=0; c5.c[2]:=5; c5.d:=1; c5.i:=1; c5.n:=2; c5.s:=true;
xsub(c3,c5,c3);
sss:=false;
end
else if (c3.d=2) and (c3.c[2]=2) and (c3.c[3]=5) then begin
xto1(a2);
exit;
end
else
if (c3.c[2]>2) or ((c3.d>1) and (c3.c[2]=2) and (c3.c[3]>=5)) then begin
c5.c[1]:=0; c5.c[2]:=5; c5.d:=1; c5.i:=1; c5.n:=2; c5.s:=true;
xsub(c5,c3,c3);
sss:=true;
end
else sss:=true;
xmul(c3,c4,c3);
xtox(c3,c2);
xto1(c3);
xtox(c3,c5);
uf:=1;
df:=1;
xtox(c2,c7);
xmul(c7,c7,c7);
repeat
xmul(c3,c7,c3);
inc(df,2);
uf:=df*(df-1);
recip;
xmul(c3,rc,c3);
xsub(c5,c3,c5);
xmul(c3,c7,c3);
inc(df,2);
uf:=df*(df-1);
recip;
xmul(c3,rc,c3);
xadd(c5,c3,c5);
until (c3.n=1) and (c3.c[1]=0);
xmul(c5,c2,c5);
xtox(c5,a2);
a2.s:=not (a1.s xor sss);
i:=a2.n;
if i>usedig then begin
dec(i,usedig);
dec(a2.d,i);
a2.n:=usedig;
end;
xelz(a2);
end;

{ the cosine uses the classic series:
cos(x) = 1 - x**2/2! + x**4/4! - x**6/6! + ... with reductions }

procedure xcos (a1: mp; var a2: mp);
begin
xtox(a1,c3);
c3.s:=true;
move(xpg,c4.c,usedig);
c4.i:=1; c4.d:=usedig-1; c4.n:=usedig; c4.s:=true;
xto2(c5);
xmul(c4,c5,c4);
xdiv(c3,c4,c3);
xfrac(c3,c3);
if c3.d>usedig then begin
c3.d:=usedig;
c3.n:=c3.i+usedig;
end;
xelz(c3);
if (c3.c[1]=0) and (c3.n=1) then begin
xto1(a2);
exit;
end;
if (c3.d=2) and (c3.c[2]=7) and (c3.c[3]=5) then begin
xto0(a2);
exit;
end
else if (c3.c[2]>7) or ((c3.d>1) and (c3.c[2]=7) and (c3.c[3]>=5))
then begin
xto1(c5);
xsub(c5,c3,c3);
sss:=true;
end
else if (c3.d=1) and (c3.c[2]=5) then begin
xtom1(a2);
exit;
end
else if c3.c[2]>=5 then begin
c5.c[1]:=0; c5.c[2]:=5; c5.d:=1; c5.i:=1; c5.n:=2; c5.s:=true;
xsub(c3,c5,c3);
sss:=false;
end
else if (c3.d=2) and (c3.c[2]=2) and (c3.c[3]=5) then begin
xto0(a2);
exit;
end
else if (c3.c[2]>2) or ((c3.d>1) and (c3.c[2]=2) and (c3.c[3]>=5))
then begin
c5.c[1]:=0; c5.c[2]:=5; c5.d:=1; c5.i:=1; c5.n:=2; c5.s:=true;
xsub(c5,c3,c3);
sss:=false;
end
else sss:=true;
xmul(c3,c4,c3);
xto1(c5);
uf:=1;
df:=0;
xtox(c3,c7);
xmul(c7,c7,c7);
xto1(c3);
repeat
xmul(c3,c7,c3);
inc(df,2);
uf:=df*(df-1);
recip;
xmul(c3,rc,c3);
xsub(c5,c3,c5);
xmul(c3,c7,c3);
inc(df,2);
uf:=df*(df-1);
recip;
xmul(c3,rc,c3);
xadd(c5,c3,c5);
until (c3.n=1) and (c3.c[1]=0);
xtox(c5,a2);
a2.s:=sss;
end;

{ tan(x) = sin(x)/cos(x) }

procedure xtan (a1: mp; var a2: mp);
begin
xsin(a1,a2);
xcos(a1,c4);
xdiv(a2,c4,a2);
if xstatus>0 then begin
xstatus:=11;
xto0(a2);
end;
end;

{ cot(x) = cos(x)/sin(x) }

procedure xcot (a1: mp; var a2: mp);
begin
xcos(a1,a2);
xsin(a1,c4);
xdiv(a2,c4,a2);
if xstatus>0 then begin
xstatus:=12;
xto0(a2);
end;
end;

{ sec(x) = 1/cos(x) }

procedure xsec (a1: mp; var a2: mp);
begin
xcos(a1,a2);
xto1(c4);
xdiv(c4,a2,a2);
if xstatus>0 then begin
xstatus:=13;
xto0(a2);
end;
end;

{ csc(x) = 1/sin(x) }

procedure xcsc (a1: mp; var a2: mp);
begin
xsin(a1,a2);
xto1(c4);
xdiv(c4,a2,a2);
if xstatus>0 then begin
xstatus:=14;
xto0(a2);
end;
end;

{ the inverse trigonometric tangent uses a series from the following two,
depending on the operand value:
arctan(x) = x - x**3/3! + x**5/5! - x**7/7! + ...
arctan(x) = (x/(x**2+1))*(1 + (2/3)*z + ((2*4)/(3*5))*z**2 +
((2*4*6)/(3*5*7))*z**3 + ... )  where z = x**2/(x**2+1),
with some reductions }

procedure xatan (a1: mp; var a2: mp);
begin
xtox(a1,c3);
xelz(c3);
if (c3.n=1) and (c3.c[1]=0) then begin
xtox(c3,a2);
exit;
end;
c3.s:=true;
inc(usedig);
xto1(c5);
if xgt(c3,c5) then begin
sss:=true;
xsub(c3,c5,c4);
xadd(c3,c5,c3);
xdiv(c4,c3,c3);
end
else sss:=false;
c5.c[1]:=0; c5.c[2]:=5; c5.d:=1; c5.n:=2; c5.i:=1; c5.s:=true;
if xgt(c5,c3) then begin
xtox(c3,c7);
uf:=1;
xmul(c3,c3,c6);
xtox(c3,c4);
xtox(c3,c2);
repeat
xmul(c4,c6,c4);
inc(uf,2);
recip;
xmul(c4,rc,c3);
xsub(c7,c3,c7);
xmul(c4,c6,c4);
inc(uf,2);
recip;
xmul(c4,rc,c3);
xadd(c7,c3,c7);
until (c3.n=1) and (c3.c[1]=0);
end
else begin
c5.c[1]:=3; c5.d:=0; c5.n:=1; c5.i:=1; c5.s:=true;
xtom1(c4);
xsqrt(c5,c5);
xmul(c3,c5,c6);
xadd(c6,c4,c6);
xadd(c3,c5,c5);
xdiv(c6,c5,c3);
c4.s:=true;
xmul(c3,c3,c5);
xadd(c5,c4,c7);
xdiv(c5,c7,c5);
xtox(c4,c6);
ef:=0;
uf:=1;
repeat
inc(ef,2);
inc(uf,2);
quot;
xmul(c4,c5,c4);
xmul(c4,rc,c4);
xadd(c6,c4,c6);
until (c4.n=1) and (c4.c[1]=0);
xmul(c6,c3,c6);
xdiv(c6,c7,c6);
move(xpg,c4.c,usedig);
c4.i:=1; c4.d:=usedig-1; c4.n:=usedig; c4.s:=true;
c5.c[1]:=6; c5.n:=1; c5.d:=0; c5.i:=1; c5.s:=true;
xdiv(c4,c5,c4);
xadd(c6,c4,c7);
end;
if sss then begin
move(xpg,c4.c,usedig);
c4.i:=1; c4.d:=usedig-1; c4.n:=usedig; c4.s:=true;
c5.c[1]:=4; c5.n:=1; c5.d:=0; c5.i:=1; c5.s:=true;
xdiv(c4,c5,c4);
xadd(c7,c4,c7);
end;
dec(usedig);
if (c7.n>usedig) and (c7.d>0) then begin
c7.n:=usedig;
c7.d:=c7.n-c7.i;
end;
xelz(c7);
xtox(c7,a2);
a2.s:=a1.s;
end;

{ arcsin(x) = arctan(x/sqrt(1-x**2)) }

procedure xasin (a1: mp; var a2: mp);
begin
xstatus:=0;
xtom1(c3);
xto1(c4);
if xgt(a1,c4) or xgt(c3,a1) then begin
xstatus:=15;
xto0(a2);
exit;
end;
if xeq(a1,c3) then begin
move(xpg,c3.c,usedig);
c3.i:=1; c3.d:=usedig-1; c3.n:=usedig; c3.s:=true;
c4.c[1]:=2;
xdiv(c3,c4,a2);
a2.s:=false;
exit;
end;
if xeq(a1,c4) then begin
move(xpg,c3.c,usedig);
c3.i:=1; c3.d:=usedig-1; c3.n:=usedig; c3.s:=true;
c4.c[1]:=2;
xdiv(c3,c4,a2);
exit;
end;
xmul(a1,a1,c3);
c3.s:=false;
xadd(c3,c4,c3);
xsqrt(c3,c3);
xdiv(a1,c3,c3);
xatan(c3,a2);
end;

{ arccos(x) = pi/2 - arcsin(x) }

procedure xacos (a1: mp; var a2: mp);
begin
xstatus:=0;
xtom1(c3);
xto1(c4);
if (xgt(a1,c4) or xgt(c3,a1)) then begin
xstatus:=16;
xto0(a2);
exit;
end;
if xeq(a1,c3) then begin
move(xpg,a2.c,usedig);
a2.i:=1; a2.d:=usedig-1; a2.n:=usedig; a2.s:=true;
exit;
end;
if xeq(a1,c4) then begin
xto0(a2);
exit;
end;
xasin(a1,a2);
move(xpg,c3.c,usedig);
c3.i:=1; c3.d:=usedig-1; c3.n:=usedig; c3.s:=true;
xto2(c4);
xdiv(c3,c4,c3);
a2.s:=(not a2.s);
xadd(c3,a2,a2);
end;

{ arccot(x) = pi/2 - arctan(x) }

procedure xacot (a1: mp; var a2: mp);
begin
xatan(a1,a2);
move(xpg,c4.c,usedig);
c4.i:=1; c4.d:=usedig-1; c4.n:=usedig; c4.s:=true;
xto2(c3);
xdiv(c4,c3,c4);
a2.s:=(not a2.s);
xadd(a2,c4,a2);
end;

{ arcsec(x) = arccos(1/x) }

procedure xasec (a1: mp; var a2: mp);
begin
xstatus:=0;
xtom1(c3);
xto1(c4);
if xgt(a1,c3) and xgt(c4,a1) then begin
xstatus:=17;
xto0(a2);
exit;
end;
xdiv(c4,a1,a2);
xacos(a2,a2);
end;

{ arccsc(x) = arcsin(1/x) }

procedure xacsc (a1: mp; var a2: mp);
begin
xstatus:=0;
xtom1(c3);
xto1(c4);
if xgt(a1,c3) and xgt(c4,a1) then begin
xstatus:=18;
xto0(a2);
exit;
end;
xdiv(c4,a1,a2);
xasin(a2,a2);
end;

{ sinh(x) = 0.5*(exp(2*x)-1)/exp(x) }

procedure xsinh(a1: mp; var a2: mp);
begin
xtox(a1,c2);
c2.s:=true;
xexp(c2,c2);
if xstatus>0 then begin
xstatus:=24;
xto0(a2);
exit;
end;
xmul(c2,c2,c3);
if xstatus>0 then begin
xstatus:=24;
xto0(a2);
exit;
end;
xto1(c7);
xsub(c3,c7,c3);
xdiv(c3,c2,c2);
if xstatus>0 then begin
xstatus:=24;
xto0(a2);
exit;
end;
c7.c[1]:=0; c7.c[2]:=5; c7.d:=1; c7.i:=1; c7.n:=2;
xmul(c2,c7,a2);
a2.s:=a1.s;
end;

{ cosh(x) = 0.5*(exp(2*x)+1)/exp(x) }

procedure xcosh (a1: mp; var a2: mp);
begin
xtox(a1,c2);
c2.s:=true;
xexp(c2,c2);
if xstatus>0 then begin
xstatus:=25;
xto0(a2);
exit;
end;
xmul(c2,c2,c3);
if xstatus>0 then begin
xstatus:=25;
xto0(a2);
exit;
end;
xto1(c7);
xadd(c3,c7,c3);
xdiv(c3,c2,c2);
if xstatus>0 then begin
xstatus:=25;
xto0(a2);
exit;
end;
c7.c[1]:=0; c7.c[2]:=5; c7.d:=1; c7.i:=1; c7.n:=2;
xmul(c2,c7,a2);
end;

{ tanh(x) = (exp(2*x)-1)/(exp(2*x)+1) }

procedure xtanh (a1: mp; var a2: mp);
begin
xtox(a1,c2);
c2.s:=true;
xexp(c2,c2);
if xstatus>0 then begin
xstatus:=0;
xto1(a2);
exit;
end;
xmul(c2,c2,c2);
if xstatus>0 then begin
xstatus:=0;
a2.c[1]:=1; a2.n:=1; a2.i:=1; a2.d:=0; a2.s:=a1.s;
exit;
end;
xto1(c7);
xsub(c2,c7,c3);
xadd(c2,c7,c4);
xdiv(c3,c4,a2);
a2.s:=a1.s;
end;

{ coth(x) = (exp(2*x)+1)/(exp(2*x)-1) }

procedure xcoth (a1: mp; var a2: mp);
begin
xtox(a1,c2);
c2.s:=true;
xelz(c2);
if (c2.c[1]=0) and (c2.n=1) then begin
xstatus:=26;
xto0(a2);
exit;
end;
xexp(c2,c2);
if xstatus>0 then begin
xstatus:=0;
a2.c[1]:=1; a2.n:=1; a2.i:=1; a2.d:=0; a2.s:=a1.s;
exit;
end;
xmul(c2,c2,c2);
if xstatus>0 then begin
xstatus:=0;
a2.c[1]:=1; a2.n:=1; a2.i:=1; a2.d:=0; a2.s:=a1.s;
exit;
end;
xto1(c7);
xadd(c2,c7,c3);
xsub(c2,c7,c4);
xdiv(c3,c4,a2);
a2.s:=a1.s;
end;

{ sech(x) = 2*exp(x)/(exp(2*x)+1) }

procedure xsech (a1: mp; var a2: mp);
begin
xtox(a1,a2);
a2.s:=true;
xexp(a2,a2);
if xstatus>0 then begin
xstatus:=0;
xto0(a2);
exit;
end;
xmul(a2,a2,c3);
xto1(c2);
xadd(c3,c2,c3);
c2.c[1]:=2;
xmul(a2,c2,a2);
xdiv(a2,c3,a2);
end;

{ csch(x) = 2*exp(x)/(exp(2*x)-1) }

procedure xcsch (a1: mp; var a2: mp);
begin
xtox(a1,a2);
xelz(a2);
if (a2.c[1]=0) and (a2.n=1) then begin
xstatus:=27;
exit;
end;
a2.s:=true;
xexp(a2,a2);
if xstatus>0 then begin
xstatus:=0;
xto0(a2);
exit;
end;
xmul(a2,a2,c3);
xtom1(c4);
xadd(c3,c4,c3);
c4.c[1]:=2; c4.s:=true;
xmul(a2,c4,a2);
xdiv(a2,c3,a2);
a2.s:=a1.s;
end;

{ arsinh(x) = ln(x+sqrt(x**2+1)) }

procedure xasinh (a1: mp; var a2: mp);
begin
xtox(a1,a2);
a2.s:=true;
xmul(a2,a2,c2);
xto1(c3);
xadd(c2,c3,c2);
xsqrt(c2,c2);
xadd(c2,a2,c2);
xln(c2,a2);
a2.s:=a1.s;
xstatus:=0;
end;

{ arcosh(x) = ln(x+sqrt(x**2-1)) }

procedure xacosh (a1: mp; var a2: mp);
begin
xto1(c3);
if xgt(c3,a1) then begin
xstatus:=28;
xto0(a2);
exit;
end;
xmul(a1,a1,c2);
c3.s:=false;
xadd(c2,c3,c2);
xsqrt(c2,c2);
xadd(a1,c2,c2);
xln(c2,a2);
end;

{ artanh(x) = 0.5*(ln(1+x)-ln(1-x)) }

procedure xatanh (a1: mp; var a2: mp);
begin
xtox(a1,a2);
xtom1(c2);
xto1(c3);
if (not (xgt(a2,c2) and xgt(c3,a2))) then begin
xstatus:=29;
xto0(a2);
exit;
end;
xadd(c3,a2,c2);
a2.s:=(not a2.s);
xadd(c3,a2,a2);
xln(c2,c2);
xln(a2,a2);
a2.s:=(not a2.s);
xadd(c2,a2,a2);
xto2(c3);
xdiv(a2,c3,a2);
end;

{ arcoth(x) = 0.5*(ln(x+1)-ln(x-1)) }

procedure xacoth (a1: mp; var a2: mp);
begin
xtox(a1,a2);
xtom1(c2);
xto1(c3);
if (not (xgt(a2,c3) or xgt(c2,a2))) then begin
xstatus:=30;
xto0(a2);
exit;
end;
a2.s:=true;
xadd(a2,c2,c2);
xadd(a2,c3,a2);
xln(c2,c2);
xln(a2,a2);
c2.s:=(not c2.s);
xadd(a2,c2,a2);
xto2(c3);
xdiv(a2,c3,a2);
if (a2.c[1]>0) or (a2.n>1) then a2.s:=a1.s;
end;

{ arsech(x) = ln((sqrt(1-x**2)+1)/x) }

procedure xasech (a1: mp; var a2: mp);
begin
xtox(a1,a2);
xto0(c2);
xto1(c3);
if (not xgt(a2,c2)) or xgt(a2,c3) then begin
xstatus:=31;
xto0(a2);
exit;
end;
xmul(a2,a2,c2);
c2.s:=(not c2.s);
xadd(c3,c2,c2);
xsqrt(c2,c2);
xadd(c2,c3,c2);
xdiv(c2,a2,a2);
xln(a2,a2);
end;

{ arcsch(x) = ln((sqrt(1+x**2)+1)/x) }

procedure xacsch (a1: mp; var a2: mp);
begin
xtox(a1,a2);
xelz(a2);
if (a2.c[1]=0) and (a2.n=1) then begin
xstatus:=32;
xto0(a2);
exit;
end;
xto1(c2);
a2.s:=true;
xmul(a2,a2,c3);
xadd(c3,c2,c3);
xsqrt(c3,c3);
xadd(c3,c2,c3);
xdiv(c3,a2,a2);
xln(a2,a2);
a2.s:=a1.s;
end;

{* interface procedures and functions *}

procedure xprecision (n: word);
begin
if n<10 then usedig:=10
else if n>maxdig then usedig:=maxdig
else usedig:=n;
end;

function xdigits: word;
begin
xdigits:=usedig;
end;

function xerror: byte;
begin
xerror:=xstatus;
end;

procedure xdis (a: mp);
begin
if a.s then write('+') else write('-');
for j:=1 to a.i do write(a.c[j]);
if a.d>0 then begin
write('.');
for j:=a.i+1 to a.n do write(a.c[j]);
end;
writeln;
end;

procedure xapp (a: mp; var b: mp; w: word);
begin
xtox(a,b);
if (a.n<=w) or (a.d=0) or (w<a.i) then exit;
i:=w;
if b.c[i+1]>4 then begin
if b.c[i]<9 then inc(b.c[i])
else begin
while (b.c[i]=9) and (i>0) do begin
b.c[i]:=0;
dec(i);
end;
if i>0 then inc(b.c[i]) else begin
b.c[w+1]:=0; b.c[1]:=1;
inc(w);
inc(b.i);
end;
end;
end;
b.d:=w-b.i;
b.n:=w;
xelz(b);
end;

procedure strtox (s: string; var a: mp; var w: word);        {uses ansistrings}
begin
if length(s)>maxbuf then begin
w:=maxbuf;
xto0(a);
exit;
end;
s1:=s;
if (s1[1]='+') or (s1[1]='-') then begin
delete(s1,1,1);
x:=1;
end
else x:=0;
if s1='' then begin
w:=1;
xto0(a);
exit;
end;
ps:=0;
ip:=0;
zs:=0;
ls:=length(s1);
while (ip=0) and (zs<ls) do begin
inc(zs);
if s1[zs]='.' then begin
if (zs=ls) and (ps=0) then ip:=ls
else if ps>0 then ip:=zs else ps:=zs;
end
else if (not (s1[zs] in ['0'..'9'])) then ip:=zs;
end;
if ip=0 then w:=0
else begin
w:=ip+x;
xto0(a);
exit;
end;
if ps=1 then begin
for zs:=2 to ls do a.c[zs]:=ord(s1[zs])-48;
a.c[1]:=0;
a.i:=1;
a.d:=ls-1;
a.n:=ls;
end
else if ps=0 then begin
for zs:=1 to ls do a.c[zs]:=ord(s1[zs])-48;
a.i:=ls;
a.d:=0;
a.n:=ls;
end
else begin
for zs:=1 to ps-1 do a.c[zs]:=ord(s1[zs])-48;
for zs:=ps+1 to ls do a.c[zs-1]:=ord(s1[zs])-48;
a.i:=ps-1;
a.d:=ls-ps;
a.n:=ls-1;
end;
if s[1]='-' then a.s:=false else a.s:=true;
xelz(a);
if (a.c[1]=0) and (a.n=1) then a.s:=true;
end;

procedure xtostr (a: mp; var s: string);                     {uses ansistrings}
begin
xtox(a,c4);
xelz(c4);
if c4.s then s:='' else s:='-';
for i:=1 to c4.i do s:=s+chr(c4.c[i]+48);
if c4.d>0 then begin
s:=s+'.';
for i:=c4.i+1 to c4.n do s:=s+chr(c4.c[i]+48);
end;
end;

{ overloaded operators }

operator = (a,b: mp) x: boolean;
begin
x:=xeq(a,b)
end;

operator > (a,b: mp) x: boolean;
begin
x:=xgt(a,b)
end;

operator < (a,b: mp) x: boolean;
begin
x:=xgt(b,a)
end;

operator >= (a,b: mp) x: boolean;
begin
x:=xgt(a,b) or xeq(a,b)
end;

operator <= (a,b: mp) x: boolean;
begin
x:=xgt(b,a) or xeq(a,b)
end;

operator := (e: ex) a: mp;
begin
a:=xfrome(e)
end;

operator := (a: mp) e: ex;
begin
e:=efromx(a)
end;

operator - (a: mp) b: mp;
begin
xtox(a,b);
xelz(b);
if (a.c[1]>0) or (a.n>1) then b.s:=(not a.s);
end;

operator + (a,b: mp) c: mp;
begin
xadd(a,b,c)
end;

operator + (a: mp; e: ex) b: mp;
begin
xadd(a,xfrome(e),b)
end;

operator + (e: ex; a: mp) b: mp;
begin
xadd(xfrome(e),a,b)
end;

operator - (a,b: mp) c: mp;
begin
xsub(a,b,c)
end;

operator - (a: mp; e: ex) b: mp;
begin
xsub(a,xfrome(e),b)
end;

operator - (e: ex; a: mp) b: mp;
begin
xsub(xfrome(e),a,b)
end;

operator * (a,b: mp) c: mp;
begin
xmul(a,b,c)
end;

operator * (a: mp; e: ex) b: mp;
begin
xmul(a,xfrome(e),b)
end;

operator * (e: ex; a: mp) b: mp;
begin
xmul(xfrome(e),a,b)
end;

operator / (a,b: mp) c: mp;
begin
xdiv(a,b,c)
end;

operator / (a: mp; e: ex) b: mp;
begin
xdiv(a,xfrome(e),b)
end;

operator / (e: ex; a: mp) b: mp;
begin
xdiv(xfrome(e),a,b)
end;

operator ** (a,b: mp) c: mp;
begin
xpow(a,b,c)
end;

operator ** (e: ex; a: mp) b: mp;
begin
xpow(xfrome(e),a,b)
end;

{ pi and Euler's constants }

function mpi : mp;
begin
move(xpg,mpi.c,usedig);
mpi.n:=usedig; mpi.i:=1; mpi.d:=usedig-1; mpi.s:=true
end;

function meu : mp;
begin
move(xeu,meu.c,usedig);
meu.n:=usedig; meu.i:=1; meu.d:=usedig-1; meu.s:=true
end;

{ mathematical functions }

function mabs (a: mp) : mp;
begin
xtox(a,mabs);
mabs.s:=true
end;

function mabs (e: ex) : mp;
begin
xtox(xfrome(e),mabs);
mabs.s:=true
end;

function mint (a: mp) : mp;
begin
xint(a,mint)
end;

function mint (e: ex) : mp;
begin
xint(xfrome(e),mint)
end;

function mfrac (a: mp) : mp;
begin
xfrac(a,mfrac)
end;

function mfrac (e: ex) : mp;
begin
xfrac(xfrome(e),mfrac)
end;

function msign (a: mp) : mp;
begin
xsign(a,msign)
end;

function msign (e: ex) : mp;
begin
xsign(xfrome(e),msign)
end;

function mceil (a: mp) : mp;
begin
xceil(a,mceil)
end;

function mceil (e: ex) : mp;
begin
xceil(xfrome(e),mceil)
end;

function mfloor (a: mp) : mp;
begin
xfloor(a,mfloor)
end;

function mfloor (e: ex) : mp;
begin
xfloor(xfrome(e),mfloor)
end;

function msqrt (a: mp) : mp;
begin
xsqrt(a,msqrt)
end;

function msqrt (e: ex) : mp;
begin
xsqrt(xfrome(e),msqrt)
end;

function msqr (a: mp) : mp;
begin
xmul(a,a,msqr)
end;

function msqr (e: ex) : mp;
begin
xtox(xfrome(e),msqr);
xmul(msqr,msqr,msqr)
end;

function mfact (a: mp) : mp;
begin
xfact(a,mfact)
end;

function mfact (e: ex) : mp;
begin
xfact(xfrome(e),mfact)
end;

function mexp (a: mp) : mp;
begin
xexp(a,mexp)
end;

function mexp (e: ex) : mp;
begin
xexp(xfrome(e),mexp)
end;

function mln (a: mp) : mp;
begin
xln(a,mln)
end;

function mln (e: ex) : mp;
begin
xln(xfrome(e),mln)
end;

function mlog (a: mp) : mp;
begin
xlog(a,mlog)
end;

function mlog (e: ex) : mp;
begin
xlog(xfrome(e),mlog)
end;

function msin (a: mp) : mp;
begin
xsin(a,msin)
end;

function msin (e: ex) : mp;
begin
xsin(xfrome(e),msin)
end;

function mcos (a: mp) : mp;
begin
xcos(a,mcos)
end;

function mcos (e: ex) : mp;
begin
xcos(xfrome(e),mcos)
end;

function mtan (a: mp) : mp;
begin
xtan(a,mtan)
end;

function mtan (e: ex) : mp;
begin
xtan(xfrome(e),mtan)
end;

function mcot (a: mp) : mp;
begin
xcot(a,mcot)
end;

function mcot (e: ex) : mp;
begin
xcot(xfrome(e),mcot)
end;

function msec (a: mp) : mp;
begin
xsec(a,msec)
end;

function msec (e: ex) : mp;
begin
xsec(xfrome(e),msec)
end;

function mcsc (a: mp) : mp;
begin
xcsc(a,mcsc)
end;

function mcsc (e: ex) : mp;
begin
xcsc(xfrome(e),mcsc)
end;

function masin (a: mp) : mp;
begin
xasin(a,masin)
end;

function masin (e: ex) : mp;
begin
xasin(xfrome(e),masin)
end;

function macos (a: mp) : mp;
begin
xacos(a,macos)
end;

function macos (e: ex) : mp;
begin
xacos(xfrome(e),macos)
end;

function matan (a: mp) : mp;
begin
xatan(a,matan)
end;

function matan (e: ex) : mp;
begin
xatan(xfrome(e),matan)
end;

function macot (a: mp) : mp;
begin
xacot(a,macot)
end;

function macot (e: ex) : mp;
begin
xacot(xfrome(e),macot)
end;

function masec (a: mp) : mp;
begin
xasec(a,masec)
end;

function masec (e: ex) : mp;
begin
xasec(xfrome(e),masec)
end;

function macsc (a: mp) : mp;
begin
xacsc(a,macsc)
end;

function macsc (e: ex) : mp;
begin
xacsc(xfrome(e),macsc)
end;

function msinh (a: mp) : mp;
begin
xsinh(a,msinh)
end;

function msinh (e: ex) : mp;
begin
xsinh(xfrome(e),msinh)
end;

function mcosh (a: mp) : mp;
begin
xcosh(a,mcosh)
end;

function mcosh (e: ex) : mp;
begin
xcosh(xfrome(e),mcosh)
end;

function mtanh (a: mp) : mp;
begin
xtanh(a,mtanh)
end;

function mtanh (e: ex) : mp;
begin
xtanh(xfrome(e),mtanh)
end;

function mcoth (a: mp) : mp;
begin
xcoth(a,mcoth)
end;

function mcoth (e: ex) : mp;
begin
xcoth(xfrome(e),mcoth)
end;

function msech (a: mp) : mp;
begin
xsech(a,msech)
end;

function msech (e: ex) : mp;
begin
xsech(xfrome(e),msech)
end;

function mcsch (a: mp) : mp;
begin
xcsch(a,mcsch)
end;

function mcsch (e: ex) : mp;
begin
xcsch(xfrome(e),mcsch)
end;

function masinh (a: mp) : mp;
begin
xasinh(a,masinh)
end;

function masinh (e: ex) : mp;
begin
xasinh(xfrome(e),masinh)
end;

function macosh (a: mp) : mp;
begin
xacosh(a,macosh)
end;

function macosh (e: ex) : mp;
begin
xacosh(xfrome(e),macosh)
end;

function matanh (a: mp) : mp;
begin
xatanh(a,matanh)
end;

function matanh (e: ex) : mp;
begin
xatanh(xfrome(e),matanh)
end;

function macoth (a: mp) : mp;
begin
xacoth(a,macoth)
end;

function macoth (e: ex) : mp;
begin
xacoth(xfrome(e),macoth)
end;

function masech (a: mp) : mp;
begin
xasech(a,masech)
end;

function masech (e: ex) : mp;
begin
xasech(xfrome(e),masech)
end;

function macsch (a: mp) : mp;
begin
xacsch(a,macsch)
end;

function macsch (e: ex) : mp;
begin
xacsch(xfrome(e),macsch)
end;

BEGIN
xprecision(75);
xstatus:=0;
rc.c[1]:=0;
rc.i:=1;
rc.s:=true;
subflag:=false;
rr:=@b5;
for i:=1 to maxbuf do rp[i]:=@rr^[i];
end.
